@@ -1,3 +1,6 @@
+from __future__ import annotations
+
+
 __copyright__ = """
 Copyright (C) 2009-2013 Andreas Kloeckner
 Copyright (C) 2020 Matt Wala
@@ -27,44 +30,66 @@ THE SOFTWARE.
 
 __doc__ = """
 Graph Algorithms
-=========================
+================
+
+.. note::
+
+    These functions are mostly geared towards directed graphs (digraphs).
 
 .. autofunction:: reverse_graph
 .. autofunction:: a_star
 .. autofunction:: compute_sccs
-.. autoclass:: CycleError
+.. autoexception:: CycleError
 .. autofunction:: compute_topological_order
 .. autofunction:: compute_transitive_closure
 .. autofunction:: contains_cycle
 .. autofunction:: compute_induced_subgraph
+.. autofunction:: as_graphviz_dot
 .. autofunction:: validate_graph
+.. autofunction:: is_connected
 
 Type Variables Used
 -------------------
 
-.. class:: T
+.. class:: NodeT
 
-    Any type.
+    Type of a graph node, can be any hashable type.
+
+.. class:: GraphT
+
+    A :class:`collections.abc.Mapping` representing a directed
+    graph. The mapping contains one key representing each node in the
+    graph, and this key maps to a :class:`collections.abc.Collection` of its
+    successor nodes. Note that most functions expect that every graph node
+    is included as a key in the graph.
 """
 
-from typing import (Collection, TypeVar, Mapping, List, Optional, Any,
-                    Callable, Set, MutableSet, Dict, Iterator, Tuple, FrozenSet)
+from typing import (
+    Any, Callable, Collection, Dict, Hashable, Iterator, List, Mapping, MutableSet,
+    Optional, Set, Tuple, TypeVar)
 
 
-T = TypeVar("T")
+try:
+    from typing import TypeAlias
+except ImportError:
+    from typing_extensions import TypeAlias
+
+
+NodeT = TypeVar("NodeT", bound=Hashable)
+
+
+GraphT: TypeAlias[NodeT] = Mapping[NodeT, Collection[NodeT]]
 
 
 # {{{ reverse_graph
 
-def reverse_graph(graph: Mapping[T, Collection[T]]) -> Dict[T, FrozenSet[T]]:
+def reverse_graph(graph: GraphT[NodeT]) -> GraphT[NodeT]:
     """
-    Reverses a graph.
+    Reverses a graph *graph*.
 
-    :param graph: A :class:`dict` representation of a directed graph, mapping each
-        node to other nodes to which it is connected by edges.
     :returns: A :class:`dict` representing *graph* with edges reversed.
     """
-    result: Dict[T, Set[T]] = {}
+    result: Dict[NodeT, Set[NodeT]] = {}
 
     for node_key, successor_nodes in graph.items():
         # Make sure every node is in the result even if it has no successors
@@ -81,10 +106,10 @@ def reverse_graph(graph: Mapping[T, Collection[T]]) -> Dict[T, FrozenSet[T]]:
 # {{{ a_star
 
 def a_star(  # pylint: disable=too-many-locals
-        initial_state: T, goal_state: T, neighbor_map: Mapping[T, Collection[T]],
-        estimate_remaining_cost: Optional[Callable[[T], float]] = None,
-        get_step_cost: Callable[[Any, T], float] = lambda x, y: 1
-        ) -> List[T]:
+        initial_state: NodeT, goal_state: NodeT, neighbor_map: GraphT[NodeT],
+        estimate_remaining_cost: Optional[Callable[[NodeT], float]] = None,
+        get_step_cost: Callable[[Any, NodeT], float] = lambda x, y: 1
+        ) -> List[NodeT]:
     """
     With the default cost and heuristic, this amounts to Dijkstra's algorithm.
     """
@@ -93,7 +118,7 @@ def a_star(  # pylint: disable=too-many-locals
 
     if estimate_remaining_cost is None:
         # pylint: disable=function-redefined
-        def estimate_remaining_cost(x: T) -> float:
+        def estimate_remaining_cost(x: NodeT) -> float:
             if x != goal_state:
                 return 1
             else:
@@ -102,7 +127,7 @@ def a_star(  # pylint: disable=too-many-locals
     class AStarNode:
         __slots__ = ["state", "parent", "path_cost"]
 
-        def __init__(self, state: T, parent: Any, path_cost: float) -> None:
+        def __init__(self, state: NodeT, parent: Any, path_cost: float) -> None:
             self.state = state
             self.parent = parent
             self.path_cost = path_cost
@@ -147,21 +172,21 @@ def a_star(  # pylint: disable=too-many-locals
 
 # {{{ compute SCCs with Tarjan's algorithm
 
-def compute_sccs(graph: Mapping[T, Collection[T]]) -> List[List[T]]:
+def compute_sccs(graph: GraphT[NodeT]) -> List[List[NodeT]]:
     to_search = set(graph.keys())
-    visit_order: Dict[T, int] = {}
+    visit_order: Dict[NodeT, int] = {}
     scc_root = {}
     sccs = []
 
     while to_search:
         top = next(iter(to_search))
-        call_stack: List[Tuple[T, Iterator[T], Optional[T]]] = [(top,
+        call_stack: List[Tuple[NodeT, Iterator[NodeT], Optional[NodeT]]] = [(top,
                                                                  iter(graph[top]),
                                                                  None)]
         visit_stack = []
         visiting = set()
 
-        scc: List[T] = []
+        scc: List[NodeT] = []
 
         while call_stack:
             top, children, last_popped_child = call_stack.pop()
@@ -214,7 +239,7 @@ class CycleError(Exception):
 
     :attr node: Node in a directed graph that is part of a cycle.
     """
-    def __init__(self, node: T) -> None:
+    def __init__(self, node: NodeT) -> None:
         self.node = node
 
 
@@ -226,7 +251,7 @@ class HeapEntry:
     Only needs to define :func:`pytools.graph.__lt__` according to
     <https://github.com/python/cpython/blob/8d21aa21f2cbc6d50aab3f420bb23be1d081dac4/Lib/heapq.py#L135-L138>.
     """
-    def __init__(self, node: T, key: Any) -> None:
+    def __init__(self, node: NodeT, key: Any) -> None:
         self.node = node
         self.key = key
 
@@ -234,15 +259,11 @@ class HeapEntry:
         return self.key < other.key
 
 
-def compute_topological_order(graph: Mapping[T, Collection[T]],
-                              key: Optional[Callable[[T], Any]] = None) -> List[T]:
+def compute_topological_order(graph: GraphT[NodeT],
+                              key: Optional[Callable[[NodeT], Any]] = None) \
+                              -> List[NodeT]:
     """Compute a topological order of nodes in a directed graph.
 
-    :arg graph: A :class:`collections.abc.Mapping` representing a directed
-        graph. The dictionary contains one key representing each node in the
-        graph, and this key maps to a :class:`collections.abc.Collection` of its
-        successor nodes.
-
     :arg key: A custom key function may be supplied to determine the order in
         break-even cases. Expects a function of one argument that is used to
         extract a comparison key from each node of the *graph*.
@@ -309,13 +330,13 @@ def compute_topological_order(graph: Mapping[T, Collection[T]],
 
 # {{{ compute transitive closure
 
-def compute_transitive_closure(graph: Mapping[T, MutableSet[T]]) -> (
-        Mapping[T, MutableSet[T]]):
+def compute_transitive_closure(
+        graph: Mapping[NodeT, MutableSet[NodeT]]) -> GraphT[NodeT]:
     """Compute the transitive closure of a directed graph using Warshall's
-        algorithm.
+    algorithm.
 
     :arg graph: A :class:`collections.abc.Mapping` representing a directed
-        graph. The dictionary contains one key representing each node in the
+        graph. The mapping contains one key representing each node in the
         graph, and this key maps to a :class:`collections.abc.MutableSet` of
         nodes that are connected to the node by outgoing edges. This graph may
         contain cycles. This object must be picklable. Every graph node must
@@ -345,14 +366,9 @@ def compute_transitive_closure(graph: Mapping[T, MutableSet[T]]) -> (
 
 # {{{ check for cycle
 
-def contains_cycle(graph: Mapping[T, Collection[T]]) -> bool:
+def contains_cycle(graph: GraphT[NodeT]) -> bool:
     """Determine whether a graph contains a cycle.
 
-    :arg graph: A :class:`collections.abc.Mapping` representing a directed
-        graph. The dictionary contains one key representing each node in the
-        graph, and this key maps to a :class:`collections.abc.Collection` of
-        nodes that are connected to the node by outgoing edges.
-
     :returns: A :class:`bool` indicating whether the graph contains a cycle.
 
     .. versionadded:: 2020.2
@@ -369,13 +385,13 @@ def contains_cycle(graph: Mapping[T, Collection[T]]) -> bool:
 
 # {{{ compute induced subgraph
 
-def compute_induced_subgraph(graph: Mapping[T, Set[T]],
-                             subgraph_nodes: Set[T]) -> Mapping[T, Set[T]]:
+def compute_induced_subgraph(graph: Mapping[NodeT, Set[NodeT]],
+                             subgraph_nodes: Set[NodeT]) -> GraphT[NodeT]:
     """Compute the induced subgraph formed by a subset of the vertices in a
-        graph.
+    graph.
 
     :arg graph: A :class:`collections.abc.Mapping` representing a directed
-        graph. The dictionary contains one key representing each node in the
+        graph. The mapping contains one key representing each node in the
         graph, and this key maps to a :class:`collections.abc.Set` of nodes
         that are connected to the node by outgoing edges.
 
@@ -397,19 +413,71 @@ def compute_induced_subgraph(graph: Mapping[T, Set[T]],
 # }}}
 
 
+# {{{ as_graphviz_dot
+
+def as_graphviz_dot(graph: GraphT[NodeT],
+                    node_labels: Optional[Callable[[NodeT], str]] = None,
+                    edge_labels: Optional[Callable[[NodeT, NodeT], str]] = None) \
+                    -> str:
+    """
+    Create a visualization of the graph *graph* in the
+    `dot <http://graphviz.org/>`__ language.
+
+    :arg node_labels: An optional function that returns node labels
+        for each node.
+
+    :arg edge_labels: An optional function that returns edge labels
+        for each pair of nodes.
+
+    :returns: A string in the `dot <http://graphviz.org/>`__ language.
+    """
+    from pytools import UniqueNameGenerator
+    id_gen = UniqueNameGenerator(forced_prefix="mynode")
+
+    from pytools.graphviz import dot_escape
+
+    if node_labels is None:
+        node_labels = lambda x: str(x)
+
+    if edge_labels is None:
+        edge_labels = lambda x, y: ""
+
+    node_to_id = {}
+
+    for node, targets in graph.items():
+        if node not in node_to_id:
+            node_to_id[node] = id_gen()
+        for t in targets:
+            if t not in node_to_id:
+                node_to_id[t] = id_gen()
+
+    # Add nodes
+    content = "\n".join(
+        [f'{node_to_id[node]} [label="{dot_escape(node_labels(node))}"];'
+         for node in node_to_id.keys()])
+
+    content += "\n"
+
+    # Add edges
+    content += "\n".join(
+        [f"{node_to_id[node]} -> {node_to_id[t]} "
+         f'[label="{dot_escape(edge_labels(node, t))}"];'
+         for (node, targets) in graph.items()
+         for t in targets])
+
+    return f"digraph mygraph {{\n{ content }\n}}\n"
+
+# }}}
+
+
 # {{{ validate graph
 
-def validate_graph(graph: Mapping[T, Collection[T]]) -> None:
+def validate_graph(graph: GraphT[NodeT]) -> None:
     """
     Validates that all successor nodes of each node in *graph* are keys in
     *graph* itself. Raises a :class:`ValueError` if not.
-
-    :arg graph: A :class:`collections.abc.Mapping` representing a directed
-        graph. The dictionary contains one key representing each node in the
-        graph, and this key maps to a :class:`collections.abc.Collection` of nodes
-        that are connected to the node by outgoing edges.
     """
-    seen_nodes: Set[T] = set()
+    seen_nodes: Set[NodeT] = set()
 
     for children in graph.values():
         seen_nodes.update(children)
@@ -420,4 +488,37 @@ def validate_graph(graph: Mapping[T, Collection[T]]) -> None:
 
 # }}}
 
+
+# {{{
+
+def is_connected(graph: GraphT[NodeT]) -> bool:
+    """
+    Returns whether all nodes in *graph* are connected, ignoring
+    the edge direction.
+
+    :returns: A :class:`bool` indicating whether the graph is connected.
+    """
+    if not graph:
+        # https://cs.stackexchange.com/questions/52815/is-a-graph-of-zero-nodes-vertices-connected
+        return True
+
+    visited = set()
+
+    undirected_graph = {node: set(children) for node, children in graph.items()}
+
+    for node, children in graph.items():
+        for child in children:
+            undirected_graph[child].add(node)
+
+    def dfs(node: NodeT) -> None:
+        visited.add(node)
+        for child in undirected_graph[node]:
+            if child not in visited:
+                dfs(child)
+
+    dfs(next(iter(graph.keys())))
+
+    return visited == graph.keys()
+
+
 # vim: foldmethod=marker
